;MICRO-READERBOARD ORNAMENT
;    DOT MATRIX VERSION
;
;LED dot matrix display: columns = anode, rows = cathode
;ATTiny2313 @ 500 kHz (internal R/C osc.)
;Fuses:  low = 0x42 (internal 4MHz, divide Clock by 8, BOD enabled)
;       high = 0x9D (BOD level 1.8V, preserve EEPROM through chip erase)
;
;PORTS:
;  PB0+1 = Column 1 (left)
;  PB2+3 = Column 2
;  PB4+5 = Column 3
;  PB6+7 = Column 4
;  PA0+1 = Column 5 (right)
;  PD0   = Row 1 (top)
;  PD1   = Row 2
;  PD2   = Row 3
;  PD3   = Row 4
;  PD4   = Row 5
;  PD5   = Row 6
;  PD6   = Row 7 (bottom)
;
;VERSION HISTORY:
;  Version 0.1     2007-01-07   Arne Rossius
;  Version 0.8     2007-01-16   Arne Rossius
;  Version 0.8.1   2007-07-07   Arne Rossius
;    * fixed bug in charset (char 0x35 = '5')
;  Version 0.8.2   2008-12-14   Arne Rossius
;    * made changes for displays with column cathode (Kingbright TC*)
;  Version 0.8.3   2010-02-10   Arne Rossius
;    * added option to select column cathode or anode
;  Version 0.8.4   2010-12-28   Arne Rossius
;    * removed extra delay in RS232 reception
;  Version 0.8.5   2011-12-27  Arne Rossius
;    * fixed charset for 'avra' backslash bug

.equ	COLUMN_CATHODE = 0 ;1 for Kingbright TC*, 0 for Kingbright TA*



.include "../../include/tn2313def.inc"

.def	scrollspeed = R10
.def	char1 = R11
.def	char2 = R12
.def	char3 = R13
.def	char4 = R14
.def	char5 = R15

.def	temp = R16
.def	temp2 = R17
.def	col1 = R18
.def	col2 = R19
.def	col3 = R20
.def	col4 = R21
.def	col5 = R22
.def	delay = R23
.def	strpos = R24
.def	charlen = R25
;R26/27 used for buffer (RAM) pointer
;R28/29 used for string (EEPROM) pointer
;R30/31 used for charset (Flash) pointer

.if (COLUMN_CATHODE == 0)
	.equ	LINE1 = 0x03
	.equ	LINE2 = 0x0C
	.equ	LINE3 = 0x30
	.equ	LINE4 = 0xC0
	.equ	LINE5 = 0x07
	.equ	LINE_OFF = 0x00
	.equ	LINE5_OFF = 0x04
.else
	.equ	LINE1 = 0xFC
	.equ	LINE2 = 0xF3
	.equ	LINE3 = 0xCF
	.equ	LINE4 = 0x3F
	.equ	LINE5 = 0x04
	.equ	LINE_OFF = 0xFF
	.equ	LINE5_OFF = 0x07
.endif
	
;===============================================================================

.dseg
	String:
		.byte 112
	RAMstack:
		.byte 16

.eseg
	.org 0x00
	.db 0xff, 8, "Hello World!   ", 0, 0

;===============================================================================

.cseg
.org 0x000
	rjmp	start
.org OVF0addr ;Timer 0 overflow interrupt
	rjmp	t0ovf
	
;===============================================================================
	
charset:
	.include "../charset.asm"
	
;===============================================================================

t0ovf:
	;Timer 0 overflow interrupt (~30 Hz)
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	;decrement delay counter
	dec	delay
	brne	t0ovf_end
	mov	delay, scrollspeed
	
	;shift display left by one column
	lsr	char1
	rol	col1
	lsr	char2
	rol	col2
	lsr	char3
	rol	col3
	lsr	char4
	rol	col4
	lsr	char5
	rol	col5
	dec	charlen
	cpi	charlen, 9 ;overflow -> next char
	brlo	t0ovf_end
nextchar:
	;load next character
	rcall	eeprom_read
	inc	YL
	mov	ZL, temp
	cpi	ZL, 0x00 ;end marker -> restart string
	breq	restart
	subi	ZL, 0x20 ;charset starts at 0x20
	clr	ZH
	lsl	ZL ;offset: multiply ZH:ZL by 6
	rol	ZH
	mov	temp, ZL
	mov	temp2, ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp
	adc	ZH, temp2
	ldi	temp, LOW(charset*2) ;add charset address
	ldi	temp2, HIGH(charset*2)
	add	ZL, temp
	adc	ZH, temp2
	lpm	charlen, Z+ ;load and invert character
	lpm	char1, Z+
	lpm	char2, Z+
	lpm	char3, Z+
	lpm	char4, Z+
	lpm	char5, Z+
.if (COLUMN_CATHODE == 0)
	com	char1
	com	char2
	com	char3
	com	char4
	com	char5
.endif
	
t0ovf_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti
	
restart:
	;restart string
	ldi	YL, 0x02
	rjmp	nextchar

;===============================================================================
	
start:
	;stack pointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;ports
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0x7B
	out	DDRD, temp
	ldi	temp, LINE5_OFF
	out	PORTA, temp
	ldi	temp, LINE_OFF
	out	PORTB, temp
	ldi	temp, 0xFF
	out	PORTD, temp
	
	;check if RS232 cable is connected
	rcall	muxwait
	sbic	PIND, 2 ;PD2 low in spite of pullup -> RS232 connected
	rjmp	play

;===============================================================================

;---------------------------------------
;         RS232 TRANSFER MODE
;---------------------------------------

transfer:
	;init registers
	ldi	XL, LOW(String)
	ldi	XH, HIGH(String)
	ldi	YL, 0x02
	
getc:
	;wait for transmission to begin
	sbis	PIND, 2
	rjmp	getc
	
	;receive byte
	rcall	waitbit ;startbit
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 0
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 1
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 2
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 3
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 4
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 5
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 6
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 7
	clc
	sbis	PIND, 2
	sec
	ror	temp
	brcs	getc_starterr ;start bit was not 1
	rcall	waitbit

	rcall	waitbit
	sbic	PIND, 2
	rjmp	getc_stoperr ;stop bit is not 0
	
	;save byte to RAM buffer
	cpi	XL, String+112 ;prevent buffer overflow
	brlo	PC+2
	ldi	XL, String+111
	st	X+, temp

	;evaluate byte
	cpi	temp, 0x03 ;CTRL-C -> go to play mode
	breq	play
	cpi	temp, 0x04 ;CTRL-D -> evaluate string
	breq	getc_done
	cpi	temp, 0x0a ;LF -> evaluate string
	breq	getc_done
	cpi	temp, 0x0d ;CR -> evaluate string
	breq	getc_done
	cpi	temp, 0x08 ;BACKSP -> remove last character and the backspace character from buffer
	breq	getc_backsp
	rjmp	getc_end
	
getc_backsp:
	cpi	XL, String+2
	brlo	getc_end
	subi	XL, 2
	
getc_starterr:
getc_stoperr:
getc_end:
	;wait until line is idle (needed if stop bit is not 0) and receive next byte
	sbic	PIND, 2
	rjmp	getc_end
	rjmp	getc
	
;--------------------
	
getc_done:
	;terminate string
	dec	XL
	ldi	temp, 0x00 ;terminate string
	st	X, temp
getc_eval:
	;evaluate received string
	cpi	XL, String ;string is empty (contains only end marker)
	brne	PC+2
	rjmp	transfer
	ldi	XL, String
	ld	temp, X
	cpi	temp, 0x1b ;ESC
	breq	getc_control
	
	;save received string to EEPROM
	ldi	YL, 0x02
getc_savestr_loop:
	ld	temp, X+
	rcall	eeprom_write
	inc	YL
	cpi	temp, 0x00
	brne	getc_savestr_loop
	rjmp	transfer
	
getc_control:
	;control string received (begins with ESC)
	inc	XL
	ld	temp, X+
	cpi	temp, 's' ;scrolling speed
	breq	getc_speed
	cpi	temp, 'S'
	breq	getc_speed
	cpi	temp, 'q' ;quit -> switch to display mode
	breq	play
	cpi	temp, 'Q'
	breq	play
	cpi	temp, 0x1b ;ESC -> switch to display mode
	breq	play
	rjmp	transfer
	
getc_speed:
	;write scrolling speed setting to eeprom
	rcall	dec2bin
	ldi	YL, 0x01
	rcall	eeprom_write
	rjmp	transfer

;===============================================================================
	
;---------------------------------------
;             "PLAY" MODE
;---------------------------------------
	
play:
	;ports
	ldi	temp, 0x7F
	out	DDRD, temp
	
	;scrolling speed
	ldi	YL, 0x01
	rcall	eeprom_read
	mov	scrollspeed, temp
	
	;initialize registers
	ldi	YL, 0x02
	clr	YH
	ldi	charlen, 0
	mov	delay, scrollspeed
	.if (COLUMN_CATHODE == 0)
	ldi	col1, 0xFF
	ldi	col2, 0xFF
	ldi	col3, 0xFF
	ldi	col4, 0xFF
	ldi	col5, 0xFF
	.else
	ldi	col1, 0x00
	ldi	col2, 0x00
	ldi	col3, 0x00
	ldi	col4, 0x00
	ldi	col5, 0x00
	.endif
	mov	char1, col1
	mov	char2, col2
	mov	char3, col3
	mov	char4, col4
	mov	char5, col5
	
	;timer 0
	ldi	temp, 0x03 ;Clk/64
	out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;enable interrupts
	sei
	
loop:
	;multiplex display
	out	PORTD, col1
	ldi	temp, LINE1
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	
	out	PORTD, col2
	ldi	temp, LINE2
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	
	out	PORTD, col3
	ldi	temp, LINE3
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	
	out	PORTD, col4
	ldi	temp, LINE4
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	
	out	PORTD, col5
	ldi	temp, LINE5
	out	PORTA, temp
	rcall	muxwait
	ldi	temp, LINE5_OFF
	out	PORTA, temp
	
	rjmp	loop
	
;--------------------

dec2bin:
	;convert 3-digit decimal number from RAM[X,X+2] to 1 byte binary
	ld	temp2, X+ ;first digit: multiply by 100 = (64 + 32 + 4)
	andi	temp2, 0x03
	lsl	temp2 ;temp2 = temp2 * 2
	mov	temp, temp2
	swap	temp ;temp = (temp2*2) * 16 = temp2 * 32
	lsl	temp2 ;temp2 = (temp2*2) * 2 = temp2 * 4
	add	temp, temp2
	swap	temp2 ;temp2 = ((temp2*2)*2) * 16 = temp2 * 64
	add	temp, temp2
	
	ld	temp2, X+ ;second digit: multiply by 10 = (8 + 2)
	andi	temp2, 0x0F
	lsl	temp2 ;temp2 = temp2 * 2
	add	temp, temp2
	lsl	temp2 ;temp2 = (temp2*2) * 4 = temp2 * 8
	lsl	temp2
	add	temp, temp2
	
	ld	temp2, X+ ;first digit
	andi	temp2, 0x0F
	add	temp, temp2
	
	ret
	
;--------------------

muxwait:
	;wait a short time
	ldi	temp, 2
	clr	temp2
	
muxwait_loop:
	dec	temp2
	brne	muxwait_loop
	dec	temp
	brne	muxwait_loop
	ret

;--------------------

waitbit:
	;wait 1/2 rs232 bit length (203 cycles + rcall + 2 cycles for saving the bit)
	ldi	temp2, 66
waitbit_loop:
	dec	temp2
	brne	waitbit_loop
	nop
	
;	rjmp PC+1 ;slow down a little (funny controller or something ...)
;	rjmp PC+1
;	rjmp PC+1
;	rjmp PC+1
;	rjmp PC+1
	
	ret

;--------------------

eeprom_write:
	;write temp to EEPROM at ZL
	sbic	EECR, EEPE
	rjmp	eeprom_write
	out	EEAR, YL
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	ret

;--------------------

eeprom_read:
	;read temp from EEPROM at YL
	sbic	EECR, EEPE
	rjmp	eeprom_read
	out	EEAR, YL
	sbi	EECR, EERE
	in	temp, EEDR
	ret
